/* AMX Mod script.
 *
 * (c) 2003, OLO
 * This file is provided as is (no warranties).
 *
 * Cvars:
 * amx_reserved_slots < value > - sets how many slots we want to keep for admins
 * amx_reservation < value > - sets kicking rules:
 * 0 - Disables checking for slots usage
 * 1 - Kick by shortes time a player without reservation when server goes full
 *     until there are so many players with reservation as amx_reserved_slots specifies.
 * 2 - Kick by worst ping a player without reservation when server goes full
 *     until there are so many players with reservation as amx_reserved_slots specifies.
 * 3 - Kick a player without reservation who tries to enter to one of the reservered slots
 *     which amount is set by amx_reserved_slots.
 */

#include <amxmod>

// Comment if you don't want to ban ppl who are
// keep reconnecting more than three times in a very
// short time even if there is no free slots
//#define BAN_RECONNECTING

#if defined BAN_RECONNECTING
new g_lastCon[32]
new g_lastConNum
new Float:g_lastCon_time
#endif

new g_logFile[16]

public plugin_init()
{
  register_plugin("Slots Reservation","0.9.4","default")
  register_cvar("amx_reserved_slots","2")
  register_cvar("amx_real_res_slots","0",FCVAR_SERVER|FCVAR_UNLOGGED)
  register_cvar("amx_reservation","1")
  get_time("admin%m%d.log",g_logFile,15)
}

new g_logDrop[] = "Reservation: ^"%s<%d><%s><>^" drop due to slot reservation"

#if defined BAN_RECONNECTING
new g_logBan[] = "Reservation: ^"%s<%d><%s><>^" ban reconnecter for 1 minute"
#endif

kickPlayer(id)
{
  // gather some info
  new name[32], authid[32]
  get_user_name(id,name,31)
  get_user_authid(id,authid,31)
  
#if defined BAN_RECONNECTING  
  // save who and how often was connecting
  new Float:gameTime = get_gametime()
  if ( !equal(g_lastCon,name) || g_lastCon_time + 5.0 < gameTime ) 
  {
    copy(g_lastCon,31,name)
    g_lastConNum = 0
  }
  else
    ++g_lastConNum
  g_lastCon_time = gameTime
#endif
  
  // if bot then kick with console cmd
  if ( is_user_bot(id) ) 
  {
    server_cmd("kick #%d", get_user_userid(id) )
    log_to_file(g_logFile,g_logDrop,name, get_user_userid(id) ,authid)
  }
#if defined BAN_RECONNECTING  
  // reconnecting player - ban him
  else if ( g_lastConNum > 3 ) 
  {
    // leave him information message
    if (g_lastConNum < 5)
    {
      client_cmd(id,"echo You have been banned for 1 minute")
      log_to_file(g_logFile,g_logBan,name, get_user_userid(id) ,authid)
    }
    
    server_cmd("banid 1 #%d", get_user_userid(id) )
    client_cmd(id,"disconnect")
  }
#endif   
  // else just disconnect him, leaving a proper message
  else 
  {
#if defined BAN_RECONNECTING
    if ( g_lastConNum > 2 )
      client_cmd(id,"echo Next time you will be banned for 1 minute")
#endif      
    client_cmd(id,"echo Dropped due to slot reservation;disconnect")
    log_to_file(g_logFile,g_logDrop,name, get_user_userid(id) ,authid)
  }
}

kickLag(){
  new who = 0, ping, loss, worst = 0
  new maxplayers = get_maxplayers()
  for(new i = 1; i <= maxplayers; ++i){
    if ( !is_user_connected(i) && !is_user_connecting(i) ) 
      continue // not used slot  
    if (get_user_flags(i)&ADMIN_RESERVATION) 
      continue // has reservation, skip him
    get_user_ping(i,ping,loss) // get ping
    if ( ping > worst ) {
      worst = ping
      who = i
    }
  }
  if (who) kickPlayer(who)
}

kickFresh(){
  new who = 0, itime, shortest = 999999999
  new maxplayers = get_maxplayers()
  for(new i = 1; i <= maxplayers; ++i){
    if ( !is_user_connected(i) && !is_user_connecting(i) )
      continue // not used slot 
    if (get_user_flags(i)&ADMIN_RESERVATION)
      continue // has reservation, skip him
    itime = get_user_time(i) // get user playing time with connection duration  
    if ( shortest > itime ) {
      shortest = itime
      who = i
    }
  }
  if (who) kickPlayer(who)
}

getSlotUsage(&playersNum,&reservedNum,&reservedMax) {
  new a = get_maxplayers()
  reservedMax = get_cvar_num("amx_reserved_slots")  
  playersNum = 0
  reservedNum = 0
  for(new i = 1; i <= a; ++i)  {
    if ( !is_user_connected(i) && !is_user_connecting(i) ) 
      continue // not used slot
    if (get_user_flags(i)&ADMIN_RESERVATION)
      ++reservedNum
    ++playersNum
  }
  return a
}

/* If there is at least one place then allow enter the game anyone.
* If the slots are reserved and not all filled then kick
* with amx_reservation rules.
*/
public client_connect(id){
  new resType = get_cvar_num( "amx_reservation" ) 
  if ( !resType )
    return PLUGIN_CONTINUE
  new reservedMax, reservedNum
  switch(checkSlots(id,reservedNum,reservedMax,resType)) {
    case 1:{
      client_cmd(id,"echo ^"You are using reserved slot (%d/%d)^"",reservedNum,reservedMax)
      /*new name[32], authid[32]
      get_user_name(id,name,31)
      get_user_authid(id,authid,31)
      log_to_file(g_logFile,"Reservation: ^"%s<%d><%s><>^" is using reserved slot (slot ^"%d^") (max ^"%d^")",
        name,get_user_userid(id),authid,reservedNum,reservedMax)*/
    }
    case 2: // this player have to be kicked - no free slots
      return PLUGIN_HANDLED // block connect in other plugins (especially in consgreet)
  }
  new realSlots = reservedMax - reservedNum
  set_cvar_num("amx_real_res_slots", (realSlots < 0) ? 0 : realSlots )
  return PLUGIN_CONTINUE
}

public client_disconnect(id){
  if ( !get_cvar_num( "amx_reservation" ) )
    return PLUGIN_CONTINUE
  new reservedMax, reservedNum, playersNum
  getSlotUsage(playersNum,reservedNum,reservedMax)
  if (get_user_flags(id)&ADMIN_RESERVATION){
    new realSlots = reservedMax - reservedNum - 1
    set_cvar_num("amx_real_res_slots", (realSlots < 0) ? 0 : realSlots )
  }
  return PLUGIN_CONTINUE
}

checkSlots(id,&reservedNum,&reservedMax,resType){
  new playersNum
  new playersMax = getSlotUsage(playersNum,reservedNum,reservedMax)
  new reservedLeft = reservedMax - reservedNum
  if ( resType < 3 ) { // This is for amx_reservation 1 and 2
    
    if (playersNum < playersMax) // Don't care until all slots are filled (f.e. 12/12)
      return 0

    if (get_user_flags(id)&ADMIN_RESERVATION) { // Connecting player is an admin
      if (reservedLeft > 0) { // And there are still free reserved slots
        switch(resType) {     // so kick one players to make server from 12/12 to 11/12
          case 1: kickFresh()
          case 2: kickLag()
        }
      }
      return 1
    }
    
    if ( reservedLeft > 0 ) {   // This is not an admin and there are still free reserved slots
      kickPlayer(id)            // We are on situation like 12/12 so kick him
      return 2                  // to make this one slot free
    }
    
  } 
  // This is for amx_reservation 3
  else if (playersNum > playersMax - reservedMax)
  {
    if (get_user_flags(id)&ADMIN_RESERVATION)
      return 1
    
    kickPlayer(id)
    return 2
  }
  
  return 0
}
